# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import partialmethod, total_ordering
from typing import Dict, List, Set
import unified_planning as up


# TODO: This features map needs to be extended with all the problem characterizations.
FEATURES = {
    "PROBLEM_CLASS": ["ACTION_BASED", "HIERARCHICAL"],
    "TIME": [
        "CONTINUOUS_TIME",
        "DISCRETE_TIME",
        "INTERMEDIATE_CONDITIONS_AND_EFFECTS",
        "TIMED_EFFECT",
        "TIMED_GOALS",
        "DURATION_INEQUALITIES",
    ],
    "EXPRESSION_DURATION": ["STATIC_FLUENTS_IN_DURATION", "FLUENTS_IN_DURATION"],
    "NUMBERS": ["CONTINUOUS_NUMBERS", "DISCRETE_NUMBERS"],
    "CONDITIONS_KIND": [
        "NEGATIVE_CONDITIONS",
        "DISJUNCTIVE_CONDITIONS",
        "EQUALITY",
        "EXISTENTIAL_CONDITIONS",
        "UNIVERSAL_CONDITIONS",
    ],
    "EFFECTS_KIND": ["CONDITIONAL_EFFECTS", "INCREASE_EFFECTS", "DECREASE_EFFECTS"],
    "TYPING": ["FLAT_TYPING", "HIERARCHICAL_TYPING"],
    "FLUENTS_TYPE": ["NUMERIC_FLUENTS", "OBJECT_FLUENTS"],
    "QUALITY_METRICS": [
        "ACTIONS_COST",
        "FINAL_VALUE",
        "MAKESPAN",
        "PLAN_LENGTH",
        "OVERSUBSCRIPTION",
    ],
    "SIMULATED_ENTITIES": ["SIMULATED_EFFECTS"],
}


class ProblemKindMeta(type):
    """Meta class used to interpret the nodehandler decorator."""

    def __new__(cls, name, bases, dct):
        def _set(self, feature, possible_features):
            assert feature in possible_features
            self._features.add(feature)

        def _has(self, feature):
            return feature in self._features

        obj = type.__new__(cls, name, bases, dct)
        for m, l in FEATURES.items():
            setattr(obj, "set_" + m.lower(), partialmethod(_set, possible_features=l))
            for f in l:
                setattr(obj, "has_" + f.lower(), partialmethod(_has, feature=f))
        return obj


@total_ordering
class ProblemKind(up.AnyBaseClass, metaclass=ProblemKindMeta):
    def __init__(self, features: Set[str] = set()):
        self._features: Set[str] = set(features)

    def __repr__(self) -> str:
        features_gen = (f"'{feature}'" for feature in self._features)
        return f'ProblemKind([{", ".join(features_gen)}])'

    def __str__(self) -> str:
        features_mapped: Dict[str, List[str]] = {}
        for k, fl in FEATURES.items():
            for feature in self._features:
                if feature in fl:
                    feature_list = features_mapped.get(k, None)
                    if feature_list is None:
                        features_mapped[k] = [feature]
                    else:
                        feature_list.append(feature)
        result_str: List[str] = [f"{k}: {fl}" for k, fl in features_mapped.items()]
        return "\n".join(result_str)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, ProblemKind):
            return self._features == oth._features
        else:
            return False

    def __hash__(self) -> int:
        res = 0
        for f in self._features:
            res += hash(f)
        return res

    def __le__(self, oth: object):
        if not isinstance(oth, ProblemKind):
            raise
        return self._features.issubset(oth._features)

    @property
    def features(self) -> Set[str]:
        return self._features

    def union(self, oth: "ProblemKind") -> "ProblemKind":
        return ProblemKind(self.features.union(oth.features))

    def intersection(self, oth: "ProblemKind") -> "ProblemKind":
        return ProblemKind(self.features.intersection(oth.features))


basic_classical_kind = ProblemKind()
basic_classical_kind.set_problem_class("ACTION_BASED")
basic_classical_kind.set_typing("FLAT_TYPING")

hierarchical_kind = ProblemKind()
hierarchical_kind.set_typing("HIERARCHICAL_TYPING")

classical_kind = ProblemKind()
classical_kind.set_problem_class("ACTION_BASED")
classical_kind.set_typing("FLAT_TYPING")
classical_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
classical_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
classical_kind.set_conditions_kind("EQUALITY")

full_classical_kind = ProblemKind()
full_classical_kind.set_problem_class("ACTION_BASED")
full_classical_kind.set_typing("FLAT_TYPING")
full_classical_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
full_classical_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
full_classical_kind.set_conditions_kind("EQUALITY")
full_classical_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
full_classical_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
full_classical_kind.set_effects_kind("CONDITIONAL_EFFECTS")

object_fluent_kind = ProblemKind()
object_fluent_kind.set_fluents_type("OBJECT_FLUENTS")

basic_numeric_kind = ProblemKind()
basic_numeric_kind.set_problem_class("ACTION_BASED")
basic_numeric_kind.set_typing("FLAT_TYPING")
basic_numeric_kind.set_numbers("DISCRETE_NUMBERS")
basic_numeric_kind.set_numbers("CONTINUOUS_NUMBERS")
basic_numeric_kind.set_fluents_type("NUMERIC_FLUENTS")

full_numeric_kind = ProblemKind()
full_numeric_kind.set_problem_class("ACTION_BASED")
full_numeric_kind.set_typing("FLAT_TYPING")
full_numeric_kind.set_numbers("DISCRETE_NUMBERS")
full_numeric_kind.set_numbers("CONTINUOUS_NUMBERS")
full_numeric_kind.set_fluents_type("NUMERIC_FLUENTS")
full_numeric_kind.set_effects_kind("INCREASE_EFFECTS")
full_numeric_kind.set_effects_kind("DECREASE_EFFECTS")

basic_temporal_kind = ProblemKind()
basic_temporal_kind.set_problem_class("ACTION_BASED")
basic_temporal_kind.set_typing("FLAT_TYPING")
basic_temporal_kind.set_time("CONTINUOUS_TIME")

temporal_kind = ProblemKind()
temporal_kind.set_problem_class("ACTION_BASED")
temporal_kind.set_typing("FLAT_TYPING")
temporal_kind.set_time("CONTINUOUS_TIME")
temporal_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
temporal_kind.set_time("TIMED_EFFECT")
temporal_kind.set_time("TIMED_GOALS")
temporal_kind.set_time("DURATION_INEQUALITIES")
temporal_kind.set_expression_duration("STATIC_FLUENTS_IN_DURATION")

quality_metrics_kind = ProblemKind()
quality_metrics_kind.set_quality_metrics("PLAN_LENGTH")
quality_metrics_kind.set_quality_metrics("ACTIONS_COST")
quality_metrics_kind.set_quality_metrics("FINAL_VALUE")

oversubscription_kind = ProblemKind()
oversubscription_kind.set_quality_metrics("OVERSUBSCRIPTION")



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import unified_planning as up
from functools import wraps
from importlib.util import find_spec
from unified_planning.environment import get_env
from unified_planning.model import ProblemKind
from unified_planning.test.pddl import enhsp


skipIf = unittest.skipIf
SkipTest = unittest.SkipTest


class skipIfEngineNotAvailable(object):
    """Skip a test if the given engine is not available."""

    def __init__(self, engine):
        self.engine = engine

    def __call__(self, test_fun):
        msg = "%s not available" % self.engine
        cond = self.engine not in get_env().factory.engines

        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)

        return wrapper


class skipIfNoOneshotPlannerForProblemKind(object):
    """Skip a test if there are no oneshot planner for the given problem kind."""

    def __init__(self, kind: ProblemKind):
        self.kind = kind

    def __call__(self, test_fun):
        msg = "no oneshot planner available for the given problem kind"
        cond = False
        try:
            get_env().factory._get_engine_class(
                "oneshot_planner", problem_kind=self.kind
            )
        except:
            cond = True

        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)

        return wrapper


class skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee(object):
    """Skip a test if there are no oneshot planner satisfies optimality guarantee."""

    def __init__(
        self, optimality_guarantee: up.engines.results.PlanGenerationResultStatus
    ):
        self.optimality_guarantee = optimality_guarantee

    def __call__(self, test_fun):
        msg = "no oneshot planner available for the given optimality guarantee"
        cond = False
        try:
            get_env().factory._get_engine_class(
                "oneshot_planner", optimality_guarantee=self.optimality_guarantee
            )
        except:
            cond = True

        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)

        return wrapper


class skipIfNoPlanValidatorForProblemKind(object):
    """Skip a test if there are no plan validator for the given problem kind."""

    def __init__(self, kind: ProblemKind):
        self.kind = kind

    def __call__(self, test_fun):
        msg = "no plan validator available for the given problem kind"
        cond = False
        try:
            get_env().factory._get_engine_class(
                "plan_validator", problem_kind=self.kind
            )
        except:
            cond = True

        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)

        return wrapper


class skipIfModuleNotInstalled(object):
    """Skip a test if there are no oneshot planner for the given problem kind."""

    def __init__(self, module_name: str):
        self.module_name = module_name

    def __call__(self, test_fun):
        msg = f"no module named {self.module_name} installed"
        test = find_spec(self.module_name)
        cond = test is None

        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)

        return wrapper


TestCase = unittest.TestCase
main = unittest.main



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from enum import Enum, auto
from typing import IO, Optional, Callable, Union


class OptimalityGuarantee(Enum):
    SATISFICING = auto()
    SOLVED_OPTIMALLY = auto()


class OneshotPlannerMixin:
    @staticmethod
    def is_oneshot_planner() -> bool:
        return True

    @staticmethod
    def satisfies(optimality_guarantee: OptimalityGuarantee) -> bool:
        return False

    def solve(
        self,
        problem: "up.model.AbstractProblem",
        callback: Optional[
            Callable[["up.engines.results.PlanGenerationResult"], None]
        ] = None,
        timeout: Optional[float] = None,
        output_stream: Optional[IO[str]] = None,
    ) -> "up.engines.results.PlanGenerationResult":
        """This method takes a up.model.AbstractProblem and returns a up.engines.results.PlanGenerationResult,
        which contains information about the solution to the problem given by the planner.

        :param problem: is the up.model.AbstractProblem to solve.
        :param callback: is a function used by the planner to give reports to the user during the problem resolution, defaults to None.
        :param timeout: is the time in seconds that the planner has at max to solve the problem, defaults to None.
        :param output_stream: is a stream of strings where the planner writes his
        output (and also errors) while the planner is solving the problem, defaults to None
        :return: the up.engines.results.PlanGenerationResult created by the planner; a data structure containing the up.plan.Plan found
        and some additional information about it.

        The only required parameter is "problem" but the planner should warn the user if callback, timeout or
        output_stream are not None and the planner ignores them."""
        assert isinstance(self, up.engines.engine.Engine)
        if not self.supports(problem.kind):
            raise up.exceptions.UPUsageError(
                f"{self.name} cannot solve this kind of problem!"
            )
        return self._solve(problem, callback, timeout, output_stream)

    def _solve(
        self,
        problem: "up.model.AbstractProblem",
        callback: Optional[
            Callable[["up.engines.results.PlanGenerationResult"], None]
        ] = None,
        timeout: Optional[float] = None,
        output_stream: Optional[IO[str]] = None,
    ) -> "up.engines.results.PlanGenerationResult":
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.engines import SequentialSimulator, SimulatorMixin
from unified_planning.model import UPCOWState
from unified_planning.test import TestCase, main
from unified_planning.test.examples import get_example_problems
from itertools import product
from typing import cast


class TestSimulator(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def simulate_on_hierarchical_blocks_world(
        self, simulator: SimulatorMixin, problem: "up.model.Problem"
    ):
        # This test takes a simulator and the problem and makes some testing.
        self.assertEqual(problem.name, "hierarchical_blocks_world")
        em = problem.env.expression_manager
        move = problem.action("move")
        clear = problem.fluent("clear")
        on = problem.fluent("on")
        ts_1 = problem.object("ts_1")
        ts_2 = problem.object("ts_2")
        ts_3 = problem.object("ts_3")
        block_1 = problem.object("block_1")
        block_2 = problem.object("block_2")
        block_3 = problem.object("block_3")
        state = UPCOWState(problem.initial_values)
        # The initial state is:
        # ts_1, block_3, block_1, block_2
        # ts_2
        # ts_3, check clear fluent in the state.
        Location = problem.user_type("Location")
        clear_check = [block_2, ts_2, ts_3]
        for o in problem.objects(Location):
            if o in clear_check:
                self.assertEqual(state.get_value(clear(o)), em.TRUE())
            else:
                self.assertEqual(state.get_value(clear(o)), em.FALSE())
        # Then we want to reach a state like this:
        # ts_1, block_1
        # ts_2, block_2
        # ts_3, block_3
        # So the moves to simulate are:
        # move(block_2, from block_1, to ts_2)
        # move(block_1, from block_3, to block_2)
        # move(block_3, from ts_1, to ts_3)
        # move(block_1, from block_2, to ts_1)
        events = simulator.get_events(move, (block_2, block_1, ts_2))
        self.assertEqual(
            len(events), 1
        )  # only 1 even corresponds to in Instantaneous Action
        state = cast(UPCOWState, simulator.apply(events[0], state))
        self.assertIsNotNone(
            state
        )  # If the state is None it means the action was not applicable

        events = simulator.get_events(move, (block_1, block_3, block_2))
        self.assertEqual(
            len(events), 1
        )  # only 1 even corresponds to in Instantaneous Action
        state = cast(UPCOWState, simulator.apply(events[0], state))
        self.assertIsNotNone(
            state
        )  # If the state is None it means the action was not applicable

        events = simulator.get_events(move, (block_3, ts_1, ts_3))
        self.assertEqual(
            len(events), 1
        )  # only 1 even corresponds to in Instantaneous Action
        state = cast(UPCOWState, simulator.apply(events[0], state))
        self.assertIsNotNone(
            state
        )  # If the state is None it means the action was not applicable

        events = simulator.get_events(move, (block_1, block_2, ts_1))
        self.assertEqual(
            len(events), 1
        )  # only 1 even corresponds to in Instantaneous Action
        state = cast(UPCOWState, simulator.apply(events[0], state))
        self.assertIsNotNone(
            state
        )  # If the state is None it means the action was not applicable
        # now we check that the state is what we desired
        Movable = problem.user_type("Movable")
        check_on = [(block_1, ts_1), (block_2, ts_2), (block_3, ts_3)]
        for obj_tuple in product(problem.objects(Movable), problem.objects(Location)):
            if obj_tuple in check_on:
                self.assertEqual(state.get_value(on(*obj_tuple)), em.TRUE())
            else:
                self.assertEqual(state.get_value(on(*obj_tuple)), em.FALSE())
        # Now we want to check if we can apply the action move (block_3, from ts_1, to ts_3),
        # which we know is not because the block_3 is not on the table space 1 (ts_1)
        event = simulator.get_events(move, (block_3, ts_1, ts_3))[0]
        self.assertFalse(simulator.is_applicable(event, state))
        # Now we check if we reached the goal
        self.assertFalse(simulator.is_goal(state))
        # To reach the goal, which is designed like this:
        # ts_1
        # ts_2
        # ts_3, block_1, block_2, block_3
        # We must:
        # move(block_3, from ts_3, to block_2)
        # move(block_1, from ts_1, to ts_3)
        # move(block_3, from block_2, to ts_1)
        # move(block_2, from ts_2, to block_1)
        # move(block_3, from ts_1, to block_2)
        # And then we check if we reached the goal.
        event = simulator.get_events(move, (block_3, ts_3, block_2))[0]
        state = cast(UPCOWState, simulator.apply(event, state))
        self.assertIsNotNone(state)
        event = simulator.get_events(move, (block_1, ts_1, ts_3))[0]
        state = cast(UPCOWState, simulator.apply(event, state))
        self.assertIsNotNone(state)
        event = simulator.get_events(move, (block_3, block_2, ts_1))[0]
        state = cast(UPCOWState, simulator.apply(event, state))
        self.assertIsNotNone(state)
        event = simulator.get_events(move, (block_2, ts_2, block_1))[0]
        state = cast(UPCOWState, simulator.apply(event, state))
        self.assertIsNotNone(state)
        event = simulator.get_events(move, (block_3, ts_1, block_2))[0]
        state = cast(UPCOWState, simulator.apply(event, state))
        self.assertIsNotNone(state)

        self.assertTrue(simulator.is_goal(state))

    def test_with_sequential_simualtor_instance(self):
        problem = self.problems["hierarchical_blocks_world"].problem
        simulator = SequentialSimulator(problem)
        self.simulate_on_hierarchical_blocks_world(simulator, problem)

    def test_with_smulator_from_factory(self):
        problem = self.problems["hierarchical_blocks_world"].problem
        with Simulator(problem) as simulator:
            self.simulate_on_hierarchical_blocks_world(simulator, problem)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from io import StringIO
import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.engines import PlanGenerationResultStatus
from unified_planning.test import TestCase, main, skipIfEngineNotAvailable
from unified_planning.test.examples import get_example_problems

VERYSMALL_TIMEOUT = 0.0001


class TestPDDLPlanner(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_basic(self):
        problem = self.problems["basic"].problem
        a = problem.action("a")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_basic_conditional(self):
        problem = self.problems["basic_conditional"].problem
        a_x = problem.action("a_x")
        a_y = problem.action("a_y")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 2)
            self.assertEqual(plan.actions[0].action, a_y)
            self.assertEqual(plan.actions[1].action, a_x)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)
            self.assertEqual(len(plan.actions[1].actual_parameters), 0)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot(self):
        problem = self.problems["robot"].problem
        move = problem.action("move")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_decrease(self):
        problem = self.problems["robot_decrease"].problem
        move = problem.action("move")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader(self):
        problem = self.problems["robot_loader"].problem
        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 4)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)
            self.assertEqual(len(plan.actions[1].actual_parameters), 1)
            self.assertEqual(len(plan.actions[2].actual_parameters), 2)
            self.assertEqual(len(plan.actions[3].actual_parameters), 1)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader_adv(self):
        problem = self.problems["robot_loader_adv"].problem
        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")

        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)
            output = StringIO()
            final_report = planner.solve(problem, output_stream=output)
            plan = final_report.plan
            planner_output = output.getvalue()
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

            self.assertIn("Domain parsed\nProblem parsed\n", planner_output)
            self.assertIn("\nGrounding..\nGrounding Time:", planner_output)
            self.assertIn("Aibr Preprocessing\n|F|:7\n", planner_output)
            self.assertIn("\n|X|:0\n|A|:15\n", planner_output)
            self.assertIn("\n|P|:0\n|E|:0\nH1 Setup Time (msec): ", planner_output)
            self.assertIn("Setting horizon to:NaN\n", planner_output)
            self.assertIn("\nRunning WA-STAR\nh(n = s_0)=3.0\n", planner_output)
            self.assertIn(
                "f(n) = 3.0 (Expanded Nodes: 0, Evaluated States: 0, Time: ",
                planner_output,
            )
            self.assertIn(
                "f(n) = 4.0 (Expanded Nodes: 2, Evaluated States: 3, Time: ",
                planner_output,
            )
            self.assertIn(
                "f(n) = 5.0 (Expanded Nodes: 5, Evaluated States: 6, Time: ",
                planner_output,
            )
            self.assertIn("Problem Solved\n\nFound Plan:\n", planner_output)
            self.assertIn(
                "\n0.0: (move l1 l2 r1)\n1.0: (load l2 r1 c1)\n2.0: (move l2 l3 r1)",
                planner_output,
            )
            self.assertIn(
                "3.0: (unload l3 r1 c1)\n4.0: (move l3 l1 r1)\n\nPlan-Length:5\n",
                planner_output,
            )
            self.assertIn(
                "\nMetric (Search):5.0\nPlanning Time (msec):", planner_output
            )
            self.assertIn("\nHeuristic Time (msec): ", planner_output)
            self.assertIn("\nSearch Time (msec): ", planner_output)
            self.assertIn("\nExpanded Nodes:7\nStates Evaluated:8\n", planner_output)
            self.assertIn(
                "\nFixed constraint violations during search (zero-crossing):0\nNumber of Dead-Ends detected:0\n",
                planner_output,
            )
            self.assertIn("\nNumber of Duplicates detected:8\n", planner_output)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader_adv_with_timeout(self):
        problem, right_plan = (
            self.problems["robot_loader_adv"].problem,
            self.problems["robot_loader_adv"].plan,
        )
        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout=VERYSMALL_TIMEOUT)
            self.assertIn(
                final_report.status,
                [
                    PlanGenerationResultStatus.TIMEOUT,
                    PlanGenerationResultStatus.SOLVED_OPTIMALLY,
                ],
            )  # It could happen that the PDDL planner manages to solve the problem
            self.assertTrue(
                final_report.plan is None or final_report.plan == right_plan
            )

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader_adv_with_long_timeout(self):
        problem = self.problems["robot_loader_adv"].problem
        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")
        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout=100)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader_adv_with_long_timeout_and_output_stream(self):
        problem = self.problems["robot_loader_adv"].problem
        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")
        output_stream = StringIO()
        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(
                problem, timeout=100, output_stream=output_stream
            )
            plan = final_report.plan
            planner_output = output_stream.getvalue()
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

            self.assertIn("Domain parsed\nProblem parsed\n", planner_output)
            self.assertIn("\nGrounding..\nGrounding Time:", planner_output)
            self.assertIn("Aibr Preprocessing\n|F|:7\n", planner_output)
            self.assertIn("\n|X|:0\n|A|:15\n", planner_output)
            self.assertIn("\n|P|:0\n|E|:0\nH1 Setup Time (msec): ", planner_output)
            self.assertIn("Setting horizon to:NaN\n", planner_output)
            self.assertIn("\nRunning WA-STAR\nh(n = s_0)=3.0\n", planner_output)
            self.assertIn(
                "f(n) = 3.0 (Expanded Nodes: 0, Evaluated States: 0, Time: ",
                planner_output,
            )
            self.assertIn(
                "f(n) = 4.0 (Expanded Nodes: 2, Evaluated States: 3, Time: ",
                planner_output,
            )
            self.assertIn(
                "f(n) = 5.0 (Expanded Nodes: 5, Evaluated States: 6, Time: ",
                planner_output,
            )
            self.assertIn("Problem Solved\n\nFound Plan:\n", planner_output)
            self.assertIn(
                "\n0.0: (move l1 l2 r1)\n1.0: (load l2 r1 c1)\n2.0: (move l2 l3 r1)",
                planner_output,
            )
            self.assertIn(
                "3.0: (unload l3 r1 c1)\n4.0: (move l3 l1 r1)\n\nPlan-Length:5\n",
                planner_output,
            )
            self.assertIn(
                "\nMetric (Search):5.0\nPlanning Time (msec):", planner_output
            )
            self.assertIn("\nHeuristic Time (msec): ", planner_output)
            self.assertIn("\nSearch Time (msec): ", planner_output)
            self.assertIn("\nExpanded Nodes:7\nStates Evaluated:8\n", planner_output)
            self.assertIn(
                "\nFixed constraint violations during search (zero-crossing):0\nNumber of Dead-Ends detected:0\n",
                planner_output,
            )
            self.assertIn("\nNumber of Duplicates detected:8\n", planner_output)

            for lm in final_report.log_messages:
                if lm.level == up.engines.LogLevel.INFO:
                    self.assertEqual(planner_output, lm.message)
                else:
                    self.assertEqual(lm.level, up.engines.LogLevel.ERROR)
                    self.assertEqual(lm.message, "")

    @skipIfEngineNotAvailable("opt-pddl-planner")
    def test_robot_loader_adv_with_short_timeout_and_output_stream(self):
        problem, right_plan = (
            self.problems["robot_loader_adv"].problem,
            self.problems["robot_loader_adv"].plan,
        )
        output_stream = StringIO()
        with OneshotPlanner(name="opt-pddl-planner") as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(
                problem, timeout=VERYSMALL_TIMEOUT, output_stream=output_stream
            )
            plan = final_report.plan
            planner_output = output_stream.getvalue()
            self.assertTrue(plan is None or plan == right_plan)

            for lm in final_report.log_messages:
                if lm.level == up.engines.LogLevel.INFO:
                    self.assertEqual(planner_output, lm.message)
                else:
                    self.assertEqual(lm.level, up.engines.LogLevel.ERROR)
                    self.assertEqual(lm.message, "")



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import inspect
import tempfile
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, skipIfEngineNotAvailable


class TestFactory(TestCase):
    @skipIfEngineNotAvailable("pyperplan")
    @skipIfEngineNotAvailable("tamer")
    def test_config_file(self):
        self.assertTrue("pyperplan" in get_env().factory.preference_list)
        with tempfile.TemporaryDirectory() as tempdir:
            config_filename = os.path.join(tempdir, "up.ini")
            with open(config_filename, "w") as config:
                config.write("[global]\n")
                config.write("engine_preference_list: tamer\n")
            env = unified_planning.environment.Environment()
            env.factory.configure_from_file(config_filename)
            self.assertTrue("pyperplan" not in env.factory.preference_list)
            self.assertEqual(env.factory.preference_list, ["tamer"])



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines some utility functions."""

from itertools import chain, combinations


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License


from unified_planning.model.metrics import *
from unified_planning.shortcuts import *
from unified_planning.engines import LogMessage, CompilationKind
from unified_planning.engines.results import LogLevel
from unified_planning.test import (
    TestCase,
    skipIfEngineNotAvailable,
    skipIfModuleNotInstalled,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.plans import ActionInstance


class TestProtobufIO(TestCase):
    @skipIfModuleNotInstalled("google.protobuf")
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()
        from unified_planning.grpc.proto_reader import ProtobufReader  # type: ignore[attr-defined]
        from unified_planning.grpc.proto_writer import ProtobufWriter  # type: ignore[attr-defined]

        self.pb_writer = ProtobufWriter()
        self.pb_reader = ProtobufReader()

    def test_fluent(self):
        problem = Problem("test")
        x = Fluent("x")

        x_pb = self.pb_writer.convert(x, problem)

        self.assertEqual(x_pb.name, "x")
        self.assertEqual(x_pb.value_type, "up:bool")

        x_up = self.pb_reader.convert(x_pb, problem)

        self.assertEqual(x_up.name, "x")
        self.assertEqual(x_up.type, BoolType())

    def test_fluent_2(self):
        problem = self.problems["robot"].problem

        for f in problem.fluents:
            f_pb = self.pb_writer.convert(f, problem)
            f_up = self.pb_reader.convert(f_pb, problem)
            self.assertEqual(f, f_up)

    def test_fluent_3(self):
        """Test to handle subtypes of usertypes of Fluent Expression"""
        problem = self.problems["hierarchical_blocks_world"].problem

        for f in problem.fluents:
            f_pb = self.pb_writer.convert(f, problem)
            f_up = self.pb_reader.convert(f_pb, problem)
            self.assertEqual(f, f_up)

    def test_objects(self):
        """Test to handle subtypes of usertypes of Fluent Expression"""
        problem = self.problems["hierarchical_blocks_world"].problem

        for o in problem.all_objects:
            o_pb = self.pb_writer.convert(o)
            o_up = self.pb_reader.convert(o_pb, problem)

            self.assertEqual(o, o_up)

    def test_expression(self):
        problem = Problem("test")
        ex = problem.env.expression_manager.true_expression

        ex_pb = self.pb_writer.convert(ex)
        ex_up = self.pb_reader.convert(ex_pb, problem)
        self.assertEqual(ex, ex_up)

        ex = problem.env.expression_manager.Int(10)

        ex_pb = self.pb_writer.convert(ex)
        ex_up = self.pb_reader.convert(ex_pb, problem)
        self.assertEqual(ex, ex_up)

    def test_fluent_expressions(self):
        problem = self.problems["hierarchical_blocks_world"].problem
        problem_pb = self.pb_writer.convert(problem)
        problem_up = self.pb_reader.convert(problem_pb)

        self.assertEqual(problem, problem_up)

    def test_type_declaration(self):
        problem = Problem("test")
        ex = UserType("object")
        ex_pb = self.pb_writer.convert(ex)
        ex_up = self.pb_reader.convert(ex_pb, problem)
        self.assertEqual(ex, ex_up)

        o = Object("o", ex)
        problem.add_object(o)

        ex = UserType("location", ex)
        ex_pb = self.pb_writer.convert(ex)
        ex_up = self.pb_reader.convert(ex_pb, problem)
        self.assertEqual(ex, ex_up)

    def test_object_declaration(self):
        problem = Problem("test")
        loc_type = UserType("location")
        obj = problem.add_object("l1", loc_type)
        obj_pb = self.pb_writer.convert(obj)
        obj_up = self.pb_reader.convert(obj_pb, problem)
        self.assertEqual(obj, obj_up)

    def test_problem(self):
        import unified_planning.grpc.generated.unified_planning_pb2 as up_pb2

        problem = self.problems["robot"].problem

        problem_pb = self.pb_writer.convert(problem)
        problem_up = self.pb_reader.convert(problem_pb)

        pb_features = set(
            [up_pb2.Feature.Name(feature) for feature in problem_pb.features]
        )
        self.assertEqual(set(problem.kind.features), pb_features)
        self.assertEqual(problem, problem_up)

    def test_action(self):
        problem = self.problems["robot"].problem
        action = problem.action("move")

        action_pb = self.pb_writer.convert(action)
        action_up = self.pb_reader.convert(action_pb, problem)

        self.assertEqual(action, action_up)

    def test_durative_action(self):
        problem = self.problems["matchcellar"].problem
        action = problem.action("mend_fuse")

        action_pb = self.pb_writer.convert(action)
        action_up = self.pb_reader.convert(action_pb, problem)

        self.assertEqual(action, action_up)

    def test_action_instance(self):
        problem = self.problems["robot"].problem
        plan = self.problems["robot"].plan
        action_instance = plan.actions[0]

        action_instance_pb = self.pb_writer.convert(action_instance)
        action_instance_up = self.pb_reader.convert(action_instance_pb, problem)

        self.assertEqual(action_instance.action, action_instance_up.action)
        self.assertEqual(
            action_instance.actual_parameters, action_instance_up.actual_parameters
        )

    def test_plan(self):
        problem = self.problems["robot"].problem
        plan = self.problems["robot"].plan

        plan_pb = self.pb_writer.convert(plan)
        plan_up = self.pb_reader.convert(plan_pb, problem)

        self.assertEqual(plan, plan_up)

    def test_time_triggered_plan(self):
        problem = self.problems["temporal_conditional"].problem
        plan = self.problems["temporal_conditional"].plan

        plan_pb = self.pb_writer.convert(plan)
        plan_up = self.pb_reader.convert(plan_pb, problem)

        self.assertEqual(plan, plan_up)

    def test_metric(self):
        problem = Problem("test")
        problem.add_quality_metric(metric=MinimizeSequentialPlanLength())
        problem.add_quality_metric(metric=MinimizeMakespan())
        problem.add_quality_metric(
            metric=MinimizeExpressionOnFinalState(
                problem.env.expression_manager.true_expression
            )
        )
        problem.add_quality_metric(
            metric=MaximizeExpressionOnFinalState(
                problem.env.expression_manager.true_expression
            )
        )

        for metric in problem.quality_metrics:
            metric_pb = self.pb_writer.convert(metric)
            metric_up = self.pb_reader.convert(metric_pb, problem)

            self.assertEqual(str(metric), str(metric_up))

    def test_log_message(self):
        def assert_log(log):
            logger_pb = self.pb_writer.convert(log)
            logger_up = self.pb_reader.convert(logger_pb)
            self.assertEqual(log, logger_up)

        log = LogMessage(LogLevel.DEBUG, "test message")
        assert_log(log)
        log = LogMessage(LogLevel.INFO, "test message")
        assert_log(log)
        log = LogMessage(LogLevel.WARNING, "test message")
        assert_log(log)
        log = LogMessage(LogLevel.ERROR, "test message")
        assert_log(log)

    @skipIfEngineNotAvailable("tamer")
    def test_plan_generation(self):
        problem = self.problems["robot"].problem

        with OneshotPlanner(name="tamer", params={"weight": 0.8}) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)

            final_report_pb = self.pb_writer.convert(final_report)
            final_report_up = self.pb_reader.convert(final_report_pb, problem)

            self.assertEqual(final_report, final_report_up)

    def test_compiler_result(self):
        problem, _ = self.problems["hierarchical_blocks_world"]
        with Compiler(name="up_grounder") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)

            ground_result_pb = self.pb_writer.convert(ground_result)
            ground_result_up = self.pb_reader.convert(ground_result_pb, problem)

            self.assertEqual(ground_result.problem, ground_result_up.problem)
            for grounded_action in ground_result.problem.actions:
                # Test both callable "map_back_action_instance" act the same on every action of the grounded_problem
                grounded_action_instance = ActionInstance(grounded_action)
                original_action_instance_up = ground_result.map_back_action_instance(
                    grounded_action_instance
                )
                original_action_instance_pb = ground_result_up.map_back_action_instance(
                    grounded_action_instance
                )
                self.assertEqual(
                    original_action_instance_pb.action,
                    original_action_instance_up.action,
                )
                self.assertEqual(
                    original_action_instance_pb.actual_parameters,
                    original_action_instance_up.actual_parameters,
                )

    @skipIfEngineNotAvailable("tamer")
    def test_validation_result(self):
        problem = self.problems["robot"].problem

        with OneshotPlanner(name="tamer", params={"weight": 0.8}) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            with PlanValidator(name="tamer") as validator:
                validation_result = validator.validate(problem, final_report.plan)

                validation_result_pb = self.pb_writer.convert(validation_result)
                validation_result_up = self.pb_reader.convert(validation_result_pb)

                self.assertEqual(validation_result, validation_result_up)


class TestProtobufProblems(TestCase):
    @skipIfModuleNotInstalled("google.protobuf")
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()
        from unified_planning.grpc.proto_reader import ProtobufReader  # type: ignore[attr-defined]
        from unified_planning.grpc.proto_writer import ProtobufWriter  # type: ignore[attr-defined]

        self.pb_writer = ProtobufWriter()
        self.pb_reader = ProtobufReader()

    def test_all_problems(self):
        for name, example in self.problems.items():
            problem = example.problem
            problem_pb = self.pb_writer.convert(problem)
            problem_up = self.pb_reader.convert(problem_pb)

            self.assertEqual(problem, problem_up)
            self.assertEqual(hash(problem), hash(problem_up))

    def test_all_plans(self):
        for name, example in self.problems.items():
            problem = example.problem
            plan = example.plan
            plan_pb = self.pb_writer.convert(plan)
            plan_up = self.pb_reader.convert(plan_pb, problem)

            self.assertEqual(plan, plan_up)
            self.assertEqual(hash(plan), hash(plan_up))

    @skipIfEngineNotAvailable("tamer")
    def test_some_plan_generations(self):
        problems = [
            "basic",
            "basic_without_negative_preconditions",
            "basic_nested_conjunctions",
            "robot_loader",
            "robot_loader_mod",
            "robot_loader_adv",
            "robot_real_constants",
            "robot_int_battery",
            "robot",
        ]
        for name in problems:
            problem = self.problems[name].problem

            with OneshotPlanner(name="tamer") as planner:
                self.assertNotEqual(planner, None)
                final_report = planner.solve(problem)

                final_report_pb = self.pb_writer.convert(final_report)
                final_report_up = self.pb_reader.convert(final_report_pb, problem)

                self.assertEqual(final_report.status, final_report_up.status)
                self.assertEqual(final_report.plan, final_report_up.plan)
                self.assertEqual(final_report.engine_name, final_report_up.engine_name)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from io import StringIO
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, skipIfEngineNotAvailable


class TestCredits(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    @skipIfEngineNotAvailable("tamer")
    def test_robot_locations_visited(self):
        credits = StringIO()
        test_credits = [
            " *** Credits ***",
            "  * In operation mode `OneshotPlanner` at line ",
            "  * Engine name: ",
            "  * Developers:  ",
            "  * Description: ",
            "  *              ",
        ]
        set_credits_stream(credits)
        with OneshotPlanner(name="tamer"):
            for test in test_credits:
                printed_credits = credits.getvalue()
                self.assertIn(test, printed_credits)
        set_credits_stream(sys.stdout)

    def test_long_env_credits(self):
        credits = StringIO()
        credits_keywords = [
            "  * Engine name: ",
            "  * Developers:  ",
            "  * Description: ",
            "  * Contacts:    ",
            "  * Website:     ",
            "  * License:     ",
        ]
        get_env().factory.print_engines_info(credits, True)
        credits_printed = credits.getvalue()
        # test that every keyword occur the same number of time in the printed result
        number_of_credits_printed = credits_printed.count(credits_keywords[0])
        if number_of_credits_printed > 0:  # If at least one credit was printed
            for keyword in credits_keywords:
                self.assertIn(keyword, credits_printed)
                self.assertEqual(
                    credits_printed.count(keyword), number_of_credits_printed
                )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import (
    classical_kind,
    full_numeric_kind,
    full_classical_kind,
)
from unified_planning.test import (
    TestCase,
    skipIfNoPlanValidatorForProblemKind,
    skipIfNoOneshotPlannerForProblemKind,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import CompilationKind
from unified_planning.engines.compilers import DisjunctiveConditionsRemover


class TestDisjunctiveConditionsRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind.union(full_numeric_kind))
    def test_robot_locations_visited(self):
        problem = self.problems["robot_locations_visited"].problem

        with Compiler(
            problem_kind=problem.kind,
            compilation_kind=CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING,
        ) as dnfr:
            res = dnfr.compile(problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING)
            dnf_problem = res.problem

            res_2 = dnfr.compile(
                problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING
            )
            dnf_problem_2 = res_2.problem

            self.assertEqual(dnf_problem, dnf_problem_2)

        self.assertEqual(len(problem.actions), 2)
        self.assertEqual(len(dnf_problem.actions), 3)

        with OneshotPlanner(problem_kind=dnf_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            dnf_plan = planner.solve(dnf_problem).plan
            plan = dnf_plan.replace_action_instances(res.map_back_action_instance)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    def test_ad_hoc_1(self):
        # mockup problem
        a = Fluent("a")
        b = Fluent("b")
        c = Fluent("c")
        d = Fluent("d")
        act = InstantaneousAction("act")
        # (a <-> (b -> c)) -> (a & d)
        # In Dnf:
        # (!a & !b) | (!a & c) | (a & b & !c) | (a & d)
        cond = Implies(Iff(a, Implies(b, c)), And(a, d))
        possible_conditions = [
            {Not(a), Not(b)},
            {Not(a), FluentExp(c)},
            {FluentExp(b), Not(c), FluentExp(a)},
            {FluentExp(a), FluentExp(d)},
        ]
        act.add_precondition(cond)
        act.add_effect(a, TRUE())
        problem = Problem("mockup")
        problem.add_fluent(a)
        problem.add_fluent(b)
        problem.add_fluent(c)
        problem.add_fluent(d)
        problem.add_action(act)
        problem.set_initial_value(a, True)
        problem.set_initial_value(b, False)
        problem.set_initial_value(c, True)
        problem.set_initial_value(d, False)
        problem.add_goal(a)
        dnfr = DisjunctiveConditionsRemover()
        res = dnfr.compile(problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING)
        dnf_problem = res.problem

        self.assertEqual(len(dnf_problem.actions), 4)
        # Cycle over all actions. For every new action assume that the precondition is equivalent
        # to one in the possible_preconditions and that no other action has the same precondition.
        for i, new_action in enumerate(dnf_problem.actions):
            self.assertEqual(new_action.effects, act.effects)
            preconditions = set(new_action.preconditions)
            self.assertIn(preconditions, possible_conditions)
            for j, new_action_oth_acts in enumerate(dnf_problem.actions):
                preconditions_oth_acts = set(new_action_oth_acts.preconditions)
                if i != j:
                    self.assertNotEqual(preconditions, preconditions_oth_acts)

    def test_ad_hoc_2(self):
        # mockup problem
        a = Fluent("a")
        act = InstantaneousAction("act")
        cond = And(a, a)
        act.add_precondition(cond)
        act.add_effect(a, TRUE())
        problem = Problem("mockup")
        problem.add_fluent(a)
        problem.add_action(act)
        problem.set_initial_value(a, True)
        problem.add_goal(a)
        dnfr = DisjunctiveConditionsRemover()
        res = dnfr.compile(problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING)
        dnf_problem = res.problem

        self.assertEqual(len(dnf_problem.actions), 1)

    def test_temproal_mockup_1(self):
        # temporal mockup
        a = Fluent("a")
        b = Fluent("b")
        c = Fluent("c")
        d = Fluent("d")
        act = DurativeAction("act")
        # !a => (b | ((c <-> d) & d))
        # In Dnf:
        # a | b | (c & d)
        exp = Implies(Not(a), Or(b, And(Iff(c, d), d)))
        act.add_condition(StartTiming(), exp)
        act.add_condition(StartTiming(1), exp)
        act.add_condition(ClosedTimeInterval(StartTiming(2), StartTiming(3)), exp)
        act.add_condition(ClosedTimeInterval(StartTiming(4), StartTiming(5)), exp)
        act.add_effect(StartTiming(6), a, TRUE())

        problem = Problem("temporal_mockup")
        problem.add_fluent(a)
        problem.add_fluent(b)
        problem.add_fluent(c)
        problem.add_fluent(d)
        problem.add_action(act)
        problem.set_initial_value(a, False)
        problem.set_initial_value(b, False)
        problem.set_initial_value(c, True)
        problem.set_initial_value(d, False)
        problem.add_goal(a)
        dnfr = DisjunctiveConditionsRemover()
        res = dnfr.compile(problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING)
        dnf_problem = res.problem
        self.assertEqual(len(dnf_problem.actions), 81)

    def test_temproal_mockup_2(self):
        # temporal mockup
        a = Fluent("a")
        b = Fluent("b")
        act = DurativeAction("act")
        exp = And(Not(a), b)
        act.add_condition(StartTiming(), exp)
        act.add_condition(StartTiming(1), exp)
        act.add_condition(ClosedTimeInterval(StartTiming(2), StartTiming(3)), exp)
        act.add_condition(ClosedTimeInterval(StartTiming(4), StartTiming(5)), exp)
        act.add_effect(StartTiming(6), a, TRUE())

        problem = Problem("temporal_mockup")
        problem.add_fluent(a)
        problem.add_fluent(b)
        problem.add_action(act)
        problem.set_initial_value(a, False)
        problem.set_initial_value(b, False)
        problem.add_goal(a)
        dnfr = DisjunctiveConditionsRemover()
        res = dnfr.compile(problem, CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING)
        dnf_problem = res.problem
        self.assertEqual(len(dnf_problem.actions), 1)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Parameter class. Both actions and fluents use this class to represent their parameters.
"""


from typing import Optional
from unified_planning.environment import Environment, get_env
import unified_planning as up


class Parameter:
    """Represents an action parameter or a fluent parameter.
    A parameter has a name, and a type."""

    def __init__(
        self,
        name: str,
        typename: "up.model.types.Type",
        env: Optional[Environment] = None,
    ):
        self._name = name
        self._typename = typename
        self._env = get_env(env)
        assert self._env.type_manager.has_type(
            typename
        ), "type of parameter does not belong to the same environment given to the parameter"

    def __repr__(self) -> str:
        return f"{str(self.type)} {self.name}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Parameter):
            return (
                self._name == oth._name
                and self._typename == oth._typename
                and self._env == oth._env
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + hash(self._typename)

    @property
    def name(self) -> str:
        """Returns the parameter name."""
        return self._name

    @property
    def type(self) -> "up.model.types.Type":
        """Returns the parameter type."""
        return self._typename

    @property
    def environment(self) -> "Environment":
        """Return the object environment"""
        return self._env

    #
    # Infix operators
    #

    def __add__(self, right):
        return self._env.expression_manager.Plus(self, right)

    def __radd__(self, left):
        return self._env.expression_manager.Plus(left, self)

    def __sub__(self, right):
        return self._env.expression_manager.Minus(self, right)

    def __rsub__(self, left):
        return self._env.expression_manager.Minus(left, self)

    def __mul__(self, right):
        return self._env.expression_manager.Times(self, right)

    def __rmul__(self, left):
        return self._env.expression_manager.Times(left, self)

    def __truediv__(self, right):
        return self._env.expression_manager.Div(self, right)

    def __rtruediv__(self, left):
        return self._env.expression_manager.Div(left, self)

    def __floordiv__(self, right):
        return self._env.expression_manager.Div(self, right)

    def __rfloordiv__(self, left):
        return self._env.expression_manager.Div(left, self)

    def __gt__(self, right):
        return self._env.expression_manager.GT(self, right)

    def __ge__(self, right):
        return self._env.expression_manager.GE(self, right)

    def __lt__(self, right):
        return self._env.expression_manager.LT(self, right)

    def __le__(self, right):
        return self._env.expression_manager.LE(self, right)

    def __pos__(self):
        return self._env.expression_manager.Plus(0, self)

    def __neg__(self):
        return self._env.expression_manager.Minus(0, self)

    def Equals(self, right):
        return self._env.expression_manager.Equals(self, right)

    def And(self, *other):
        return self._env.expression_manager.And(self, *other)

    def __and__(self, *other):
        return self._env.expression_manager.And(self, *other)

    def __rand__(self, *other):
        return self._env.expression_manager.And(*other, self)

    def Or(self, *other):
        return self._env.expression_manager.Or(self, *other)

    def __or__(self, *other):
        return self._env.expression_manager.Or(self, *other)

    def __ror__(self, *other):
        return self._env.expression_manager.Or(*other, self)

    def Not(self):
        return self._env.expression_manager.Not(self)

    def __invert__(self):
        return self._env.expression_manager.Not(self)

    def Xor(self, *other):
        em = self._env.expression_manager
        return em.And(em.Or(self, *other), em.Not(em.And(self, *other)))

    def __xor__(self, *other):
        em = self._env.expression_manager
        return em.And(em.Or(self, *other), em.Not(em.And(self, *other)))

    def __rxor__(self, other):
        em = self._env.expression_manager
        return em.And(em.Or(*other, self), em.Not(em.And(*other, self)))

    def Implies(self, right):
        return self._env.expression_manager.Implies(self, right)

    def Iff(self, right):
        return self._env.expression_manager.Iff(self, right)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the dnf remover class."""

import unified_planning as up
import unified_planning.engines as engines
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.compilers.utils import get_fresh_name, replace_action
from unified_planning.engines.results import CompilerResult
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model import (
    FNode,
    Problem,
    InstantaneousAction,
    DurativeAction,
    TimeInterval,
    Timing,
    Action,
    ProblemKind,
)
from unified_planning.model.walkers import Dnf
from typing import List, Tuple, Dict, cast
from itertools import product
from functools import partial


class DisjunctiveConditionsRemover(engines.engine.Engine, CompilerMixin):
    """DisjunctiveConditions remover class: this class offers the capability
    to transform a problem with preconditions not in the DNF form
    into one with all the preconditions in DNF form (an OR of AND);
    Then this OR is decomposed into subactions, therefore after this
    remover is called, every action condition or precondition will be
    an AND of leaf nodes.
    """

    @property
    def name(self):
        return "dcrm"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= DisjunctiveConditionsRemover.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.DISJUNCTIVE_CONDITIONS_REMOVING

    def _compile(
        self,
        problem: "up.model.AbstractProblem",
        compilation_kind: "up.engines.CompilationKind",
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        env = problem.env

        new_to_old: Dict[Action, Action] = {}

        new_problem = problem.clone()
        new_problem.name = f"{self.name}_{problem.name}"
        new_problem.clear_actions()

        dnf = Dnf(env)
        for a in problem.actions:
            if isinstance(a, InstantaneousAction):
                new_precond = dnf.get_dnf_expression(
                    env.expression_manager.And(a.preconditions)
                )
                if new_precond.is_or():
                    for and_exp in new_precond.args:
                        na = self._create_new_action_with_given_precond(
                            new_problem, and_exp, a
                        )
                        new_to_old[na] = a
                        new_problem.add_action(na)
                else:
                    na = self._create_new_action_with_given_precond(
                        new_problem, new_precond, a
                    )
                    new_to_old[na] = a
                    new_problem.add_action(na)
            elif isinstance(a, DurativeAction):
                interval_list: List[TimeInterval] = []
                conditions: List[List[FNode]] = []
                # save the timing, calculate the dnf of the and of all the conditions at the same time
                # and then save it in conditions.
                # conditions contains lists of Fnodes, where [a,b,c] means a or b or c
                for i, cl in a.conditions.items():
                    interval_list.append(i)
                    new_cond = dnf.get_dnf_expression(env.expression_manager.And(cl))
                    if new_cond.is_or():
                        conditions.append(new_cond.args)
                    else:
                        conditions.append([new_cond])
                conditions_tuple = cast(Tuple[List[FNode], ...], product(*conditions))
                for cond_list in conditions_tuple:
                    nda = self._create_new_durative_action_with_given_conds_at_given_times(
                        new_problem, interval_list, cond_list, a
                    )
                    new_to_old[nda] = a
                    new_problem.add_action(nda)
            else:
                raise NotImplementedError

        return CompilerResult(
            new_problem, partial(replace_action, map=new_to_old), self.name
        )

    def _create_new_durative_action_with_given_conds_at_given_times(
        self,
        new_problem,
        interval_list: List[TimeInterval],
        cond_list: List[FNode],
        original_action: DurativeAction,
    ) -> DurativeAction:
        new_action = original_action.clone()
        new_action.name = get_fresh_name(new_problem, original_action.name)
        new_action.clear_conditions()
        for i, c in zip(interval_list, cond_list):
            if c.is_and():
                for co in c.args:
                    new_action.add_condition(i, co)
            else:
                new_action.add_condition(i, c)
        return new_action

    def _create_new_action_with_given_precond(
        self, new_problem, precond: FNode, original_action: InstantaneousAction
    ) -> InstantaneousAction:
        new_action = original_action.clone()
        new_action.name = get_fresh_name(new_problem, original_action.name)
        new_action.clear_preconditions()
        if precond.is_and():
            for leaf in precond.args:
                new_action.add_precondition(leaf)
        else:
            new_action.add_precondition(precond)
        return new_action



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
import re
import sys
import unified_planning as up
import unified_planning.environment
import unified_planning.model.walkers as walkers
from unified_planning.model import (
    DurativeAction,
    InstantaneousAction,
    Fluent,
    Parameter,
    Object,
)
from unified_planning.model.types import _UserType, _RealType, _IntType
from typing import IO, Dict, List, Optional, cast, Union
from io import StringIO

ANML_KEYWORDS = {
    "action",
    "and",
    "constant",
    "duration",
    "else",
    "fact",
    "fluent",
    "function",
    "goal",
    "in",
    "instance",
    "motivated",
    "predicate",
    "symbol",
    "variable",
    "when",
    "with",
    "decomposition",
    "use",
    "coincident",
    "comprise",
    "comprises",
    "contain",
    "contains",
    "exists",
    "forall",
    "implies",
    "iff",
    "not",
    "or",
    "ordered",
    "unordered",
    "xor",
    "UNDEFINED",
    "all",
    "end",
    "false",
    "infinity",
    "object",
    "start",
    "true",
    "boolean",
    "float",
    "rational",
    "integer",
    "string",
    "type",
    "set",
    "subset",
    "powerset",
    "intersect",
    "union",
    "elt",
}

# The following map is used to mangle the invalid names by their class.
INITIAL_LETTER: Dict[type, str] = {
    InstantaneousAction: "a",
    DurativeAction: "a",
    Fluent: "f",
    Parameter: "p",
    Object: "o",
}


class ConverterToANMLString(walkers.DagWalker):
    """Expression converter to an ANML string."""

    def __init__(
        self,
        names_mapping: Dict[
            Union[
                "up.model.Type",
                "up.model.Action",
                "up.model.Parameter",
                "up.model.Fluent",
                "up.model.Object",
            ],
            str,
        ],
        env: "up.environment.Environment",
    ):
        walkers.DagWalker.__init__(self)
        self._names_mapping = names_mapping
        self.simplifier = env.simplifier

    def convert(self, expression):
        """Converts the given expression to a ANML string."""
        return self.walk(
            self.simplifier.simplify(expression)
        )  # NOTE maybe the converter could remove the first and last char, if they are '(' and ')'

    def walk_exists(self, expression, args):
        assert len(args) == 1
        vars_string_gen = (
            f"{_get_anml_name(v.type, self._names_mapping)} {_get_anml_name(v, self._names_mapping)}"
            for v in expression.variables()
        )
        return f'(exists({", ".join(vars_string_gen)}) {{ {args[0]} }})'

    def walk_forall(self, expression, args):
        assert len(args) == 1
        vars_string_gen = (
            f"{_get_anml_name(v.type, self._names_mapping)} {_get_anml_name(v, self._names_mapping)}"
            for v in expression.variables()
        )
        return f'(forall({", ".join(vars_string_gen)}) {{ {args[0]} }})'

    def walk_variable_exp(self, expression, args):
        assert len(args) == 0
        return _get_anml_name(expression.variable(), self._names_mapping)

    def walk_and(self, expression, args):
        assert len(args) > 1
        return f'({" and ".join(args)})'

    def walk_or(self, expression, args):
        assert len(args) > 1
        return f'({" or ".join(args)})'

    def walk_not(self, expression, args):
        assert len(args) == 1
        return f"(not {args[0]})"

    def walk_implies(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} implies {args[1]})"

    def walk_iff(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} iff {args[1]})"

    def walk_fluent_exp(self, expression, args):
        if len(args) == 0:
            return self._names_mapping[expression.fluent()]
        else:
            return f'{self._names_mapping[expression.fluent()]}({", ".join(args)})'

    def walk_param_exp(self, expression, args):
        assert len(args) == 0
        return _get_anml_name(expression.parameter(), self._names_mapping)

    def walk_object_exp(self, expression, args):
        assert len(args) == 0
        return _get_anml_name(expression.object(), self._names_mapping)

    def walk_bool_constant(self, expression, args):
        assert len(args) == 0
        if expression.bool_constant_value():
            return "true"
        return "false"

    def walk_real_constant(self, expression, args):
        assert len(args) == 0
        frac = cast(Fraction, expression.constant_value())
        return f"({frac.numerator}/{frac.denominator})"

    def walk_int_constant(self, expression, args):
        assert len(args) == 0
        return str(expression.constant_value())

    def walk_plus(self, expression, args):
        assert len(args) > 1
        return f"({' + '.join(args)})"

    def walk_minus(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} - {args[1]})"

    def walk_times(self, expression, args):
        assert len(args) > 1
        return f"({' * '.join(args)})"

    def walk_div(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} / {args[1]})"

    def walk_le(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} <= {args[1]})"

    def walk_lt(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} < {args[1]})"

    def walk_equals(self, expression, args):
        assert len(args) == 2
        return f"({args[0]} == {args[1]})"


class ANMLWriter:
    """This class can be used to write a Problem in ANML."""

    def __init__(self, problem: "up.model.Problem"):
        self.problem = problem

    def _write_problem(self, out: IO[str]):
        names_mapping: Dict[
            Union[
                "up.model.Type",
                "up.model.Action",
                "up.model.Parameter",
                "up.model.Fluent",
                "up.model.Object",
            ],
            str,
        ] = {}
        # Init names_mapping.
        names_mapping[self.problem.env.type_manager.BoolType()] = "boolean"
        names_mapping[self.problem.env.type_manager.IntType()] = "integer"
        names_mapping[self.problem.env.type_manager.RealType()] = "float"
        for t in self.problem.user_types:
            ut = cast(_UserType, t)
            if _is_valid_anml_name(ut.name):  # No renaming needed
                names_mapping[t] = ut.name
        for a in self.problem.actions:
            if _is_valid_anml_name(a.name):  # No renaming needed
                names_mapping[a] = a.name
        for f in self.problem.fluents:
            if _is_valid_anml_name(f.name):  # No renaming needed
                names_mapping[f] = f.name
        for o in self.problem.all_objects:
            if _is_valid_anml_name(o.name):  # No renaming needed
                names_mapping[o] = o.name

        for t in self.problem.user_types:
            anml_type_name = _get_anml_name(t, names_mapping)
            out.write(f"type {anml_type_name}")
            if cast(_UserType, t).father is None:
                out.write(";\n")
            else:
                # For construction in the Problem, the father of a UserType is always added before the UserType itself.
                father = cast(_UserType, t).father
                assert father is not None
                assert names_mapping[father] is not None
                out.write(f" < {names_mapping[father]};\n")
        static_fluents = self.problem.get_static_fluents()
        for f in self.problem.fluents:
            parameters = [
                f"{_get_anml_name(ap.type, names_mapping)} {_get_anml_name(ap, names_mapping)}"
                for ap in f.signature
            ]
            params_written = f'({", ".join(parameters)})' if len(parameters) > 0 else ""
            if f in static_fluents:
                out.write(
                    f"constant {_get_anml_name(f.type, names_mapping)} {_get_anml_name(f, names_mapping)}{params_written};\n"
                )
            else:
                out.write(
                    f"fluent {_get_anml_name(f.type, names_mapping)} {_get_anml_name(f, names_mapping)}{params_written};\n"
                )

        converter = ConverterToANMLString(names_mapping, self.problem.env)

        for a in self.problem.actions:
            if isinstance(a, up.model.InstantaneousAction):
                parameters = [
                    f"{_get_anml_name(ap.type, names_mapping)} {_get_anml_name(ap, names_mapping)}"
                    for ap in a.parameters
                ]
                out.write(
                    f'action {_get_anml_name(a, names_mapping)}({", ".join(parameters)}) {{\n'
                )
                for p in a.preconditions:
                    out.write(f"   [ start ] {converter.convert(p)};\n")
                for e in a.effects:
                    out.write(f"   {self._convert_effect(e, converter, None, 3)}")
                out.write("};\n")
            elif isinstance(a, DurativeAction):
                parameters = [
                    f"{_get_anml_name(ap.type, names_mapping)} {_get_anml_name(ap, names_mapping)}"
                    for ap in a.parameters
                ]
                out.write(
                    f'action {_get_anml_name(a, names_mapping)}({", ".join(parameters)}) {{\n'
                )
                left_bound = " > " if a.duration.is_left_open() else " >= "
                right_bound = " < " if a.duration.is_right_open() else " <= "
                out.write(
                    f"   duration{left_bound}{converter.convert(a.duration.lower)} and "
                )
                out.write(
                    f"duration{right_bound}{converter.convert(a.duration.upper)};\n"
                )
                for i, cl in a.conditions.items():
                    for c in cl:
                        out.write(
                            f"   {self._convert_anml_interval(i)} {converter.convert(c)};\n"
                        )
                for ti, el in a.effects.items():
                    for e in el:
                        out.write(f"   {self._convert_effect(e, converter, ti, 3)}")
                out.write("};\n")
            else:
                raise NotImplementedError

        for t in self.problem.user_types:  # Define objects
            obj_names = [
                _get_anml_name(o, names_mapping)
                for o in self.problem.objects(t)
                if o.type == t
            ]
            if len(obj_names) > 0:
                out.write(
                    f'instance {_get_anml_name(t, names_mapping)} {", ".join(obj_names)};\n'
                )

        for fe, v in self.problem.initial_values.items():
            assert fe.is_fluent_exp()
            if fe.fluent() in static_fluents:
                out.write(f"{converter.convert(fe)} := {converter.convert(v)};\n")
            else:
                out.write(
                    f"[ start ] {converter.convert(fe)} := {converter.convert(v)};\n"
                )

        for ti, el in self.problem.timed_effects.items():
            for e in el:
                out.write(self._convert_effect(e, converter, ti))

        for g in self.problem.goals:
            out.write(f"[ end ] {converter.convert(g)};\n")

        for i, gl in self.problem.timed_goals.items():
            for g in gl:
                out.write(f"{self._convert_anml_interval(i)} {converter.convert(g)};\n")

    def print_problem(self):
        """Prints to std output the ANML problem."""
        self._write_problem(sys.stdout)

    def get_problem(self) -> str:
        """Returns the ANML problem."""
        out = StringIO()
        self._write_problem(out)
        return out.getvalue()

    def write_problem(self, filename: str):
        """Dumps to file the ANML problem."""
        with open(filename, "w") as f:
            self._write_problem(f)

    def _convert_effect(
        self,
        effect: "up.model.Effect",
        converter: ConverterToANMLString,
        timing: "up.model.Timing" = None,
        spaces_from_left: int = 0,
    ) -> str:
        results: List[str] = []
        anml_timing = (
            self._convert_anml_timing(timing) if timing is not None else "start"
        )
        if effect.is_conditional():
            results.append(
                f'when [ {anml_timing} ] {converter.convert(effect.condition)}\n{spaces_from_left*" "}{{'
            )
        results.append(f"[ {anml_timing} ] ")
        results.append(converter.convert(effect.fluent))
        if effect.is_assignment():
            results.append(" := ")
        elif effect.is_increase():
            results.append(f" :increase ")
        elif effect.is_decrease():
            results.append(f" :decrease ")
        else:
            raise NotImplementedError
        results.append(f"{converter.convert(effect.value)};\n")
        if effect.is_conditional():
            results.append(f'{spaces_from_left*" "}}}\n')
        return "".join(results)

    def _convert_anml_timing(self, timing: "up.model.Timing") -> str:
        time = "start" if timing.is_from_start() else "end"
        if timing.delay > 0:
            return f"{time} + {str(timing.delay)}"
        elif timing.delay == 0:
            return time
        else:  # timing.delay < 0
            return f"{time} - {str(timing.delay * (-1))}"

    def _convert_anml_interval(self, interval: "up.model.TimeInterval") -> str:
        left_bracket = "(" if interval.is_left_open() else "["
        right_bracket = ")" if interval.is_left_open() else "]"
        return f"{left_bracket} {self._convert_anml_timing(interval.lower)}, {self._convert_anml_timing(interval.upper)} {right_bracket}"


def _is_valid_anml_name(name: str) -> bool:
    regex = re.compile(r"^[a-zA-Z][a-zA-Z0-9_]*")
    if (
        re.match(regex, name) is None or name in ANML_KEYWORDS
    ):  # If the name does not start with an alphabetic char or is a keyword
        return False
    return True


def _get_anml_valid_name(
    item: Union[
        "up.model.Type",
        "up.model.Action",
        "up.model.Parameter",
        "up.model.Fluent",
        "up.model.Object",
    ]
) -> str:
    """This function returns a valid ANML name."""
    if isinstance(item, up.model.Type):
        assert item.is_user_type()
        name = cast(_UserType, item).name
    else:
        name = item.name
    regex = re.compile(r"^[a-zA-Z]+.*")
    if (
        re.match(regex, name) is None
    ):  # If the name does not start with an alphabetic char, we make it start with one.
        name = f'{INITIAL_LETTER.get(type(item), "x")}_{name}'
    name = re.sub("[^0-9a-zA-Z_]", "_", name)  # Substitute non-valid elements with "_"
    while (
        name in ANML_KEYWORDS
    ):  # If the name is in the keywords, apply an underscore at the end until it is not a keyword anymore.
        name = f"{name}_"
    return name


def _get_anml_name(
    item: Union[
        "up.model.Type",
        "up.model.Action",
        "up.model.Parameter",
        "up.model.Fluent",
        "up.model.Object",
    ],
    names_mapping: Dict[
        Union[
            "up.model.Type",
            "up.model.Action",
            "up.model.Parameter",
            "up.model.Fluent",
            "up.model.Object",
        ],
        str,
    ],
) -> str:
    """Important note: This method updates the names_mapping"""
    new_name: Optional[str] = names_mapping.get(item, None)
    if new_name is None:  # The type is not in the dictionary, so his name must be added
        if isinstance(item, up.model.Type) and item.is_int_type():
            num_type = cast(_IntType, item)
            left_bound = (
                "(-infinity"
                if num_type.lower_bound is None
                else f"[{str(num_type.lower_bound)}"
            )
            right_bound = (
                "infinity)"
                if num_type.upper_bound is None
                else f"{str(num_type.upper_bound)}]"
            )
            new_name = f"integer {left_bound}, {right_bound}"
        elif isinstance(item, up.model.Type) and item.is_real_type():
            num_real_type = cast(_RealType, item)
            if num_real_type.lower_bound is None:
                left_bound = "(-infinity"
            elif num_real_type.lower_bound.denominator == 1:
                left_bound = f"[{str(num_real_type.lower_bound)}.0"
            else:
                left_bound = f"[{str(num_real_type.lower_bound)}"
            if num_real_type.upper_bound is None:
                right_bound = "infinity)"
            elif num_real_type.upper_bound.denominator == 1:
                right_bound = f"{str(num_real_type.upper_bound)}.0]"
            else:
                right_bound = f"{str(num_real_type.upper_bound)}]"
            new_name = f"float {left_bound}, {right_bound}"
        else:  # We mangle the name and get a fresh one
            new_name = _get_anml_valid_name(item)
            test_name = new_name  # Init values
            count = 0
            while (
                test_name in names_mapping.values()
            ):  # Loop until we find a fresh name
                test_name = f"{new_name}_{str(count)}"
                count += 1
            new_name = test_name
            assert _is_valid_anml_name(new_name)
        names_mapping[
            item
        ] = new_name  # Once a fresh valid name is found, update the map.
    return cast(str, new_name)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase


class TestInfixNotation(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    def test_infix_with_exp(self):
        i_1 = FluentExp(Fluent("i_1", IntType()))
        r_1 = FluentExp(Fluent("r_1", RealType()))
        i_2 = FluentExp(Fluent("i_2", IntType()))
        r_2 = FluentExp(Fluent("r_2", RealType()))
        b_1 = FluentExp(Fluent("b_1"))
        b_2 = Fluent("b_2")
        b_3 = FluentExp(Fluent("b_3"))
        self._test_helper_function(i_1, r_1, i_2, r_2, b_1, b_2, b_3)

    def test_infix_with_fluents(self):
        i_1 = Fluent("i_1", IntType())
        r_1 = Fluent("r_1", RealType())
        i_2 = Fluent("i_2", IntType())
        r_2 = Fluent("r_2", RealType())
        b_1 = Fluent("b_1")
        b_2 = Fluent("b_2")
        b_3 = Fluent("b_3")
        self._test_helper_function(i_1, r_1, i_2, r_2, b_1, b_2, b_3)

    def test_infix_with_variables(self):
        i_1 = Variable("i_1", IntType())
        r_1 = Variable("r_1", RealType())
        i_2 = Variable("i_2", IntType())
        r_2 = Variable("r_2", RealType())
        b_1 = Variable("b_1", BoolType())
        b_2 = Variable("b_2", BoolType())
        b_3 = Variable("b_3", BoolType())
        self._test_helper_function(i_1, r_1, i_2, r_2, b_1, b_2, b_3)

    def test_infix_with_parameters(self):
        env = get_env()
        i_1 = Parameter("i_1", IntType(), env)
        r_1 = Parameter("r_1", RealType(), env)
        i_2 = Parameter("i_2", IntType(), env)
        r_2 = Parameter("r_2", RealType(), env)
        b_1 = Parameter("b_1", BoolType(), env)
        b_2 = Parameter("b_2", BoolType(), env)
        b_3 = Parameter("b_3", BoolType(), env)
        self._test_helper_function(i_1, r_1, i_2, r_2, b_1, b_2, b_3)

    def test_infix_on_objects(self):
        ut = UserType("ut")
        o_1 = Object("o_1", ut)
        o_2 = Object("o_2", ut)
        f_i = Fluent("f_i", ut, id=IntType())
        f_o = Fluent("f_o", ut)
        expressions: List[Tuple[FNode, FNode]] = [
            (Equals(o_1, o_2), o_1.Equals(o_2)),
            (Equals(o_1, f_i(1)), o_1.Equals(f_i(1))),
            (Equals(o_1, f_o), o_1.Equals(f_o)),
        ]
        for exp, tested_exp in expressions:
            self.assertEqual(exp, tested_exp)

    def _test_helper_function(self, i_1, r_1, i_2, r_2, b_1, b_2, b_3):
        expressions: List[Tuple[FNode, FNode]] = [
            (Plus(i_1, i_2), i_1 + i_2),
            (Plus(i_1, 1), i_1 + 1),
            (Plus(2, i_2), 2 + i_2),
            (Plus(r_1, r_2), r_1 + r_2),
            (Plus(r_1, 1), r_1 + 1),
            (Plus(2, r_2), 2 + r_2),
            (Minus(i_1, i_2), i_1 - i_2),
            (Minus(i_1, 1), i_1 - 1),
            (Minus(2, i_2), 2 - i_2),
            (Times(i_1, i_2), i_1 * i_2),
            (Times(i_1, 1), i_1 * 1),
            (Times(2, i_2), 2 * i_2),
            (Div(i_1, i_2), i_1 / i_2),
            (Div(i_1, 1), i_1 / 1),
            (Div(2, i_2), 2 / i_2),
            (Div(i_1, i_2), i_1 // i_2),
            (Div(i_1, 1), i_1 // 1),
            (Div(2, i_2), 2 // i_2),
            (GT(i_1, i_2), i_1 > i_2),
            (GE(i_1, 1), i_1 >= 1),
            (LT(2, i_2), 2 < i_2),
            (LE(i_1, i_2), i_1 <= i_2),
            (Plus(0, i_1), +i_1),
            (Minus(0, i_1), -i_1),
            (Equals(i_1, 1), i_1.Equals(1)),
            (Equals(r_1, r_2), r_1.Equals(r_2)),
            (And(b_1, b_2), b_1.And(b_2)),
            (And([b_1, b_2, b_3]), b_1.And([b_2, b_3])),
            (And([b_1, b_2, b_3]), b_1.And(b_2, b_3)),
            (And(b_1, b_2), b_1 & b_2),
            (Or(b_1, b_2), b_1.Or(b_2)),
            (Or([b_1, b_2, b_3]), b_1.Or([b_2, b_3])),
            (Or([b_1, b_2, b_3]), b_1.Or(b_2, b_3)),
            (Or(b_1, b_2), b_1 | b_2),
            (Not(b_1), b_1.Not()),
            (Not(b_1), ~b_1),
            (And(Or(b_1, b_2), Not(And(b_1, b_2))), b_1.Xor(b_2)),
            (And(Or(b_1, b_2, b_3), Not(And(b_1, b_2, b_3))), b_1.Xor([b_2, b_3])),
            (And(Or(b_1, b_2, b_3), Not(And(b_1, b_2, b_3))), b_1.Xor(b_2, b_3)),
            (And(Or(b_1, b_2), Not(And(b_1, b_2))), b_1 ^ b_2),
            (Implies(b_1, b_2), b_1.Implies(b_2)),
            (Iff(b_1, b_2), b_1.Iff(b_2)),
        ]
        for exp, tested_exp in expressions:
            self.assertEqual(exp, tested_exp)



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import unified_planning as up
import unified_planning.engines.mixins as mixins
import unified_planning.engines.results
from unified_planning.model import ProblemKind
from unified_planning.engines.engine import Engine
from unified_planning.engines.meta_engine import MetaEngine
from unified_planning.engines.results import (
    PlanGenerationResultStatus,
    PlanGenerationResult,
)
from unified_planning.engines.mixins.oneshot_planner import OptimalityGuarantee
from unified_planning.utils import powerset
from typing import Type, IO, Callable, Optional, Union, List, Tuple
from fractions import Fraction


class OversubscriptionPlanner(MetaEngine, mixins.OneshotPlannerMixin):
    def __init__(self, *args, **kwargs):
        MetaEngine.__init__(self, *args, **kwargs)
        mixins.OneshotPlannerMixin.__init__(self)

    @property
    def name(self) -> str:
        return f"OversubscriptionPlanner[{self.engine.name}]"

    @staticmethod
    def satisfies(optimality_guarantee: OptimalityGuarantee) -> bool:
        if optimality_guarantee == OptimalityGuarantee.SATISFICING:
            return True
        return False

    @staticmethod
    def is_compatible_engine(engine: Type[Engine]) -> bool:
        return engine.is_oneshot_planner()  # type: ignore

    @staticmethod
    def _supported_kind(engine: Type[Engine]) -> "ProblemKind":
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        supported_kind.set_expression_duration("STATIC_FLUENTS_IN_DURATION")
        supported_kind.set_expression_duration("FLUENTS_IN_DURATION")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        final_supported_kind = supported_kind.intersection(engine.supported_kind())
        final_supported_kind.set_quality_metrics("OVERSUBSCRIPTION")
        return final_supported_kind

    @staticmethod
    def _supports(problem_kind: "ProblemKind", engine: Type[Engine]) -> bool:
        return problem_kind <= OversubscriptionPlanner._supported_kind(engine)

    def _solve(
        self,
        problem: "up.model.AbstractProblem",
        callback: Optional[Callable[["PlanGenerationResult"], None]] = None,
        timeout: Optional[float] = None,
        output_stream: Optional[IO[str]] = None,
    ) -> "PlanGenerationResult":
        assert isinstance(problem, up.model.Problem)
        assert isinstance(self.engine, mixins.OneshotPlannerMixin)
        if len(problem.quality_metrics) == 0:
            goals: List[Tuple["up.model.FNode", Union[Fraction, int]]] = []
        else:
            assert len(problem.quality_metrics) == 1
            qm = problem.quality_metrics[0]
            assert isinstance(qm, up.model.metrics.Oversubscription)
            goals = list(qm.goals.items())
        q = []
        for l in powerset(goals):
            cost: Union[Fraction, int] = 0
            sg = []
            for g, c in l:
                cost += c
                sg.append(g)
            q.append((cost, sg))
        q.sort(reverse=True, key=lambda t: t[0])
        incomplete = False
        for t in q:
            new_problem = problem.clone()
            new_problem.clear_quality_metrics()
            for g in t[1]:
                new_problem.add_goal(g)
            start = time.time()
            res = self.engine.solve(new_problem, callback, timeout, output_stream)
            if timeout is not None:
                timeout -= min(timeout, time.time() - start)
            if res.status in up.engines.results.POSITIVE_OUTCOMES:
                if incomplete:
                    status = PlanGenerationResultStatus.SOLVED_SATISFICING
                else:
                    status = PlanGenerationResultStatus.SOLVED_OPTIMALLY
                return PlanGenerationResult(
                    status,
                    res.plan,
                    self.name,
                    metrics=res.metrics,
                    log_messages=res.log_messages,
                )
            elif res.status == PlanGenerationResultStatus.TIMEOUT:
                return PlanGenerationResult(
                    PlanGenerationResultStatus.TIMEOUT, None, self.name
                )
            elif res.status in [
                PlanGenerationResultStatus.MEMOUT,
                PlanGenerationResultStatus.INTERNAL_ERROR,
                PlanGenerationResultStatus.UNSUPPORTED_PROBLEM,
                PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY,
            ]:
                incomplete = True
        if incomplete:
            status = PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY
        else:
            status = PlanGenerationResultStatus.UNSOLVABLE_PROVEN
        return PlanGenerationResult(status, None, self.name)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class handles:
    def __init__(self, *what):
        self.what = what

    def __call__(self, func):
        func._what = self.what
        return func


class Converter:
    def __init__(self):
        self.functions = {}
        for k in dir(self):
            v = getattr(self, k)
            if hasattr(v, "_what"):
                for x in v._what:
                    self.functions[x] = v

    def convert(self, element, *args):
        f = self.functions[type(element)]
        return f(element, *args)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction
from collections import OrderedDict
from typing import List, Optional, Set, Union
import unified_planning as up
import unified_planning.environment
import unified_planning.model.walkers as walkers
from unified_planning.model.fnode import FNode
import unified_planning.model.operators as op


class Simplifier(walkers.dag.DagWalker):
    """Performs basic simplifications of the input expression."""

    def __init__(
        self,
        env: "unified_planning.environment.Environment",
        problem: Optional["unified_planning.model.problem.Problem"] = None,
    ):
        walkers.dag.DagWalker.__init__(self)
        self.env = env
        self.manager = env.expression_manager
        if problem is not None:
            self.static_fluents = problem.get_static_fluents()
        else:
            self.static_fluents = set()
        self.problem: Optional["unified_planning.model.problem.Problem"] = problem

    def _number_to_fnode(self, value: Union[int, float, Fraction]) -> FNode:
        if isinstance(value, int):
            fnode = self.manager.Int(value)
        else:
            fnode = self.manager.Real(Fraction(value))
        return fnode

    def simplify(self, expression: FNode) -> FNode:
        """Performs basic simplification of the given expression.

        If a problem is given at the costructor, it also uses the static fluents of the problem for a better simplification."""
        return self.walk(expression)

    def walk_and(self, expression: FNode, args: List[FNode]) -> FNode:
        if len(args) == 2 and args[0] == args[1]:
            return args[0]

        new_args: OrderedDict[FNode, bool] = OrderedDict()
        for a in args:
            if a.is_true():
                continue
            if a.is_false():
                return self.manager.FALSE()
            if a.is_and():
                for s in a.args:
                    if self.walk_not(self.manager.Not(s), [s]) in new_args:
                        return self.manager.FALSE()
                    new_args[s] = True
            else:
                if self.walk_not(self.manager.Not(a), [a]) in new_args:
                    return self.manager.FALSE()
                new_args[a] = True

        if len(new_args) == 0:
            return self.manager.TRUE()
        elif len(new_args) == 1:
            return next(iter(new_args))
        else:
            return self.manager.And(new_args.keys())

    def walk_or(self, expression: FNode, args: List[FNode]) -> FNode:
        if len(args) == 2 and args[0] == args[1]:
            return args[0]

        new_args: OrderedDict[FNode, bool] = OrderedDict()
        for a in args:
            if a.is_false():
                continue
            if a.is_true():
                return self.manager.TRUE()
            if a.is_or():
                for s in a.args:
                    if self.walk_not(self.manager.Not(s), [s]) in new_args:
                        return self.manager.TRUE()
                    new_args[s] = True
            else:
                if self.walk_not(self.manager.Not(a), [a]) in new_args:
                    return self.manager.TRUE()
                new_args[a] = True

        if len(new_args) == 0:
            return self.manager.FALSE()
        elif len(new_args) == 1:
            return next(iter(new_args))
        else:
            return self.manager.Or(new_args.keys())

    def walk_not(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 1
        child = args[0]
        if child.is_bool_constant():
            l = child.bool_constant_value()
            return self.manager.Bool(not l)
        elif child.is_not():
            return child.arg(0)

        return self.manager.Not(child)

    def walk_iff(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_bool_constant() and sr.is_bool_constant():
            l = sl.bool_constant_value()
            r = sr.bool_constant_value()
            return self.manager.Bool(l == r)
        elif sl.is_bool_constant():
            if sl.bool_constant_value():
                return sr
            else:
                return self.manager.Not(sr)
        elif sr.is_bool_constant():
            if sr.bool_constant_value():
                return sl
            else:
                return self.manager.Not(sl)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Iff(sl, sr)

    def walk_implies(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_bool_constant():
            l = sl.bool_constant_value()
            if l:
                return sr
            else:
                return self.manager.TRUE()
        elif sr.is_bool_constant():
            r = sr.bool_constant_value()
            if r:
                return self.manager.TRUE()
            else:
                return self.manager.Not(sl)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Implies(sl, sr)

    def walk_exists(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 1
        free_vars: Set[
            "up.model.variable.Variable"
        ] = self.env.free_vars_oracle.get_free_variables(args[0])
        vars = tuple(var for var in expression.variables() if var in free_vars)
        if len(vars) == 0:
            return args[0]
        return self.manager.Exists(args[0], *vars)

    def walk_forall(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 1
        free_vars: Set[
            "up.model.variable.Variable"
        ] = self.env.free_vars_oracle.get_free_variables(args[0])
        vars = tuple(var for var in expression.variables() if var in free_vars)
        if len(vars) == 0:
            return args[0]
        return self.manager.Forall(args[0], *vars)

    def walk_equals(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l == r)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Equals(sl, sr)

    def walk_le(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l <= r)
        return self.manager.LE(sl, sr)

    def walk_lt(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l < r)
        return self.manager.LT(sl, sr)

    def walk_fluent_exp(self, expression: FNode, args: List[FNode]) -> FNode:
        if expression.fluent() not in self.static_fluents:
            return self.manager.FluentExp(expression.fluent(), tuple(args))
        else:
            assert self.problem is not None
            for a in args:
                if not a.is_constant():
                    return self.manager.FluentExp(expression.fluent(), tuple(args))
            return self.problem.initial_value(
                self.manager.FluentExp(expression.fluent(), tuple(args))
            )

    def walk_plus(self, expression: FNode, args: List[FNode]) -> FNode:
        new_args_plus: List[FNode] = list()
        accumulator: Union[int, Fraction] = 0
        # divide constant FNode and accumulate their value into accumulator
        for a in args:
            if a.is_int_constant() or a.is_real_constant():
                accumulator += a.constant_value()
            elif a.is_plus():
                for s in a.args:
                    if s.is_int_constant() or s.is_real_constant():
                        accumulator += s.constant_value()
                    else:
                        new_args_plus.append(s)
            else:
                new_args_plus.append(a)
        # if accumulator != 0 create it as a constant FNode and then add all the non-constant FNodes found
        # else return 0 or all the non-constant FNodes found
        if accumulator != 0:
            fnode_acc = self.manager.Plus(
                *new_args_plus, self._number_to_fnode(accumulator)
            )
            return fnode_acc
        else:
            if len(new_args_plus) == 0:
                return self.manager.Int(0)
            else:
                return self.manager.Plus(new_args_plus)

    def walk_minus(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2
        left, right = args
        value: Union[Fraction, int] = 0
        if (left.is_int_constant() or left.is_real_constant()) and (
            right.is_int_constant() or right.is_real_constant()
        ):
            value = left.constant_value() - right.constant_value()
            fnode_constant_values = self._number_to_fnode(value)
            return fnode_constant_values
        elif right.is_int_constant() or right.is_real_constant():
            if right.constant_value() < 0:
                value = -right.constant_value()
                fnode_constant_values = self._number_to_fnode(value)
                return self.manager.Plus(left, fnode_constant_values)
            else:
                return self.manager.Minus(left, right)
        else:
            return self.manager.Minus(left, right)

    def walk_times(self, expression: FNode, args: List[FNode]) -> FNode:
        new_args_times: List[FNode] = list()
        accumulator: Union[int, Fraction] = 1
        # divide constant FNode and accumulate their value into accumulator
        for a in args:
            if a.is_int_constant() or a.is_real_constant():
                if a.constant_value() == 0:
                    return self.manager.Int(0)
                else:
                    accumulator *= a.constant_value()
            elif a.is_times():
                for s in a.args:
                    if s.is_int_constant() or s.is_real_constant():
                        if s.constant_value() == 0:
                            return self.manager.Int(0)
                        else:
                            accumulator *= s.constant_value()
                    else:
                        new_args_times.append(s)
            else:
                new_args_times.append(a)
        # if accumulator != 1 create it as a constant FNode and then add all the non-constant FNodes found
        # else return  or all the non-constant FNodes found
        if accumulator != 1:
            fnode_acc = self._number_to_fnode(accumulator)
            return self.manager.Times(*new_args_times, fnode_acc)
        else:
            if len(new_args_times) == 0:
                return self.manager.Int(1)
            else:
                return self.manager.Times(new_args_times)

    def walk_div(self, expression: FNode, args: List[FNode]) -> FNode:
        assert len(args) == 2
        left, right = args
        value: Union[Fraction, int, float] = 0
        if left.is_int_constant() and right.is_int_constant():
            if (left.constant_value() % right.constant_value()) == 0:
                value = int(left.constant_value() / right.constant_value())
            else:
                value = Fraction(left.constant_value(), right.constant_value())
        elif (left.is_int_constant() or left.is_real_constant()) and (
            right.is_int_constant() or right.is_real_constant()
        ):
            assert right.constant_value() != 0
            value = Fraction(left.constant_value(), right.constant_value())
        else:
            return self.manager.Div(left, right)
        return self._number_to_fnode(value)

    @walkers.handles(op.CONSTANTS)
    @walkers.handles(
        op.OperatorKind.PARAM_EXP,
        op.OperatorKind.VARIABLE_EXP,
        op.OperatorKind.OBJECT_EXP,
        op.OperatorKind.TIMING_EXP,
    )
    def walk_identity(self, expression: FNode, args: List[FNode]) -> FNode:
        return expression



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the meta engine interface."""

from unified_planning.exceptions import UPUsageError
from unified_planning.engines.engine import Engine, EngineMeta
from unified_planning.model import ProblemKind
from functools import partial
from typing import Type


class MetaEngineMeta(EngineMeta):
    def __getitem__(self, engine_class: Type[Engine]):
        assert issubclass(self, MetaEngine)
        if not self.is_compatible_engine(engine_class):  # type: ignore
            raise UPUsageError(f"{engine_class.name} is not compatible with the meta engine {self.name}")  # type: ignore

        class MetaEngineImpl(self):  # type: ignore
            _engine_class = engine_class

            @staticmethod
            def supported_kind() -> ProblemKind:
                return self._supported_kind(engine_class)  # type: ignore

            @staticmethod
            def supports(problem_kind: ProblemKind) -> bool:
                return self._supports(problem_kind, engine_class)  # type: ignore

        return MetaEngineImpl


class MetaEngine(Engine, metaclass=MetaEngineMeta):
    """This class represents a meta engine.
    A meta engine is an engine that can be instantiated over a generic engine.
    e.g. OversubscriptionPlanner[Tamer] is an Engine class that use the MetaEngine
    OversubscriptionPlanner over the Tamer Engine."""

    def __init__(self, *args, **kwargs):
        self._engine = self._engine_class(*args, **kwargs)

    @property
    def engine(self) -> Engine:
        """Returns the engine used by this MetaEngine class"""
        return self._engine

    @staticmethod
    def is_compatible_engine(engine: Type[Engine]) -> bool:
        """Returns true iff the given engine is compatible with this meta engine"""
        raise NotImplementedError

    @staticmethod
    def _supported_kind(engine: Type[Engine]) -> ProblemKind:
        """Returns the supported kind of this meta engine with the given engine"""
        raise NotImplementedError

    @staticmethod
    def _supports(problem_kind: ProblemKind, engine: Type[Engine]) -> bool:
        """Returns true iff the given problem kind is supported by this meta
        engine with the given engine"""
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Iterator, List, Optional, Tuple, Union
import unified_planning as up
from unified_planning.exceptions import UPUsageError


class Event:
    """This is an abstract class representing an event."""

    @property
    def conditions(self) -> List["up.model.FNode"]:
        raise NotImplementedError

    @property
    def effects(self) -> List["up.model.Effect"]:
        raise NotImplementedError

    @property
    def simulated_effect(self) -> Optional["up.model.SimulatedEffect"]:
        raise NotImplementedError


class SimulatorMixin:
    """
    SimulatorMixin abstract class.
    This class defines the interface that a simulator must implement.

    Important NOTE: The AbstractProblem instance is given at the constructor.
    """

    def __init__(self, problem: "up.model.AbstractProblem") -> None:
        """
        Takes an instance of a problem and eventually some parameters, that represent
        some specific settings of the SimulatorMixin.
        :param problem: the problem that defines the domain in which the simulation exists.
        """
        self._problem = problem
        self_class = type(self)
        assert issubclass(
            self_class, up.engines.engine.Engine
        ), "SimulatorMixin does not implement the up.engines.Engine class"
        if not self_class.supports(problem.kind):
            raise UPUsageError(
                f"The problem named: {problem.name} is not supported by the {self_class}."
            )

    def is_applicable(self, event: "Event", state: "up.model.ROState") -> bool:
        """
        Returns True if the given event conditions are evaluated as True in the given state;
        returns False otherwise.
        :param state: the state where the event conditions are checked.
        :param event: the event whose conditions are checked.
        :return: Whether or not the event is applicable in the given state.
        """
        return self._is_applicable(event, state)

    def _is_applicable(self, event: "Event", state: "up.model.ROState") -> bool:
        return (
            len(self.get_unsatisfied_conditions(event, state, early_termination=True))
            == 0
        )

    def get_unsatisfied_conditions(
        self, event: "Event", state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        """
        Returns the list of unsatisfied event conditions evaluated in the given state.
        If the flag "early_termination" is set, the method ends and returns at the first unsatisfied condition.
        :param state: The State in which the event conditions are evaluated.
        :param early_termination: Flag deciding if the method ends and returns at the first unsatisfied condition.
        :return: The list of all the event conditions that evaluated to False or the list containing the first condition evaluated to False if the flag "early_termination" is set.
        """
        return self._get_unsatisfied_conditions(
            event, state, early_termination=early_termination
        )

    def _get_unsatisfied_conditions(
        self, event: "Event", state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        raise NotImplementedError

    def apply(
        self, event: "Event", state: "up.model.COWState"
    ) -> Optional["up.model.COWState"]:
        """
        Returns None if the event is not applicable in the given state, otherwise returns a new COWState,
        which is a copy of the given state but the applicable effects of the event are applied; therefore
        some fluent values are updated.
        :param state: the state where the event formulas are calculated.
        :param event: the event that has the information about the conditions to check and the effects to apply.
        :return: None if the event is not applicable in the given state, a new COWState with some updated values
         if the event is applicable.
        """
        return self._apply(event, state)

    def _apply(
        self, event: "Event", state: "up.model.COWState"
    ) -> Optional["up.model.COWState"]:
        raise NotImplementedError

    def apply_unsafe(
        self, event: "Event", state: "up.model.COWState"
    ) -> "up.model.COWState":
        """
        Returns a new COWState, which is a copy of the given state but the applicable effects of the event are applied; therefore
        some fluent values are updated.
        IMPORTANT NOTE: Assumes that self.is_applicable(state, event) returns True
        :param state: the state where the event formulas are evaluated.
        :param event: the event that has the information about the effects to apply.
        :return: A new COWState with some updated values.
        """
        return self._apply_unsafe(event, state)

    def _apply_unsafe(
        self, event: "Event", state: "up.model.COWState"
    ) -> "up.model.COWState":
        raise NotImplementedError

    def get_applicable_events(self, state: "up.model.ROState") -> Iterator["Event"]:
        """
        Returns a view over all the events that are applicable in the given State;
        an Event is considered applicable in a given State, when all the Event condition
        simplify as True when evaluated in the State.
        :param state: the state where the formulas are evaluated.
        :return: an Iterator of applicable Events.
        """
        return self._get_applicable_events(state)

    def _get_applicable_events(self, state: "up.model.ROState") -> Iterator["Event"]:
        raise NotImplementedError

    def get_events(
        self,
        action: "up.model.Action",
        parameters: Union[
            Tuple["up.model.Expression", ...], List["up.model.Expression"]
        ],
    ) -> List["Event"]:
        """
        Returns a list containing all the events derived from the given action, grounded with the given parameters.
        :param action: the action containing the information to create the event.
        :param parameters: the parameters needed to ground the action
        :return: the List of Events derived from this action with these parameters.
        """
        if len(action.parameters) != len(parameters):
            raise UPUsageError(
                "The parameters given action do not have the same length of the given parameters."
            )
        return self._get_events(action, parameters)

    def _get_events(
        self,
        action: "up.model.Action",
        parameters: Union[
            Tuple["up.model.Expression", ...], List["up.model.Expression"]
        ],
    ) -> List["Event"]:
        raise NotImplementedError

    @staticmethod
    def is_simulator():
        return True

    def is_goal(self, state: "up.model.ROState") -> bool:
        """
        Returns True if the given state satisfies the goals of the problem.
        :param state: the State in which the problem goals are evaluated.
        :return: True if the evaluation of every goal is True, False otherwise.
        """
        return self._is_goal(state)

    def _is_goal(self, state: "up.model.ROState") -> bool:
        return len(self.get_unsatisfied_goals(state, early_termination=True)) == 0

    def get_unsatisfied_goals(
        self, state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        """
        Returns the list of unsatisfied goals evaluated in the given state.
        If the flag "early_termination" is set, the method ends and returns at the first unsatisfied goal.
        :param state: The State in which the problem goals are evaluated.
        :param early_termination: Flag deciding if the method ends and returns at the first unsatisfied goal.
        :return: The list of all the goals that evaluated to False or the list containing the first goal evaluated to False if the flag "early_termination" is set.
        """
        return self._get_unsatisfied_goals(state, early_termination)

    def _get_unsatisfied_goals(
        self, state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from unified_planning.model.types import _UserType
from unified_planning.exceptions import UPProblemDefinitionError, UPValueError
from typing import Iterator, List, Union, Optional, cast


class ObjectsSetMixin:
    """
    This class is a mixin that contains a set of objects with some related methods.

    NOTE: when this mixin is used in combination with other mixins that share some
    of the attributes (e.g. env, add_user_type_method, has_name_method), it is required
    to pass the very same arguments to the mixins constructors.
    """

    def __init__(self, env, add_user_type_method, has_name_method):
        self._env = env
        self._add_user_type_method = add_user_type_method
        self._has_name_method = has_name_method
        self._objects: List["up.model.object.Object"] = []

    @property
    def env(self) -> "up.environment.Environment":
        """Returns the problem environment."""
        return self._env

    def add_object(
        self,
        obj_or_name: Union["up.model.object.Object", str],
        typename: Optional["up.model.types.Type"] = None,
    ) -> "up.model.object.Object":
        """Add the given object to the problem, constructing it from the parameters if needed.

        :param obj_or_name: Either an Object instance or a string containing the name of the object.
        :param typename: If the first argument contains only the name of the object, this parameter should contain
                         its type, to allow creating the object.
        :return: The Object that was passed or constructed.

        Examples
        --------
        >>> from unified_planning.shortcuts import *
        >>> problem = Problem()
        >>> cup = UserType("Cup")
        >>> o1 = Object("o1", cup)  # creates a new object o1
        >>> problem.add_object(o1)  # adds it to the problem
        o1
        >>> o2 = problem.add_object("o2", cup)  # alternative syntax to create a new object and add it to the problem.
        """
        if isinstance(obj_or_name, up.model.object.Object):
            assert typename is None
            obj = obj_or_name
            assert (
                obj.environment == self._env
            ), "Object does not have the same environemt fo the problem"
        else:
            assert typename is not None, "Missing type of the object"
            obj = up.model.object.Object(obj_or_name, typename, self._env)
        if self._has_name_method(obj.name):
            raise UPProblemDefinitionError("Name " + obj.name + " already defined!")
        self._objects.append(obj)
        if obj.type.is_user_type():
            self._add_user_type_method(obj.type)
        return obj

    def add_objects(self, objs: List["up.model.object.Object"]):
        """Adds the given objects."""
        for obj in objs:
            self.add_object(obj)

    def object(self, name: str) -> "up.model.object.Object":
        """Returns the object with the given name."""
        for o in self._objects:
            if o.name == name:
                return o
        raise UPValueError(f"Object of name: {name} is not defined!")

    def has_object(self, name: str) -> bool:
        """Returns true if the object with the given name is in the problem."""
        for o in self._objects:
            if o.name == name:
                return True
        return False

    def objects(
        self, typename: "up.model.types.Type"
    ) -> Iterator["up.model.object.Object"]:
        """Returns the objects of the given user type and of its heirs."""
        for obj in self._objects:
            if cast(_UserType, obj.type).is_subtype(typename):
                yield obj

    @property
    def all_objects(self) -> List["up.model.object.Object"]:
        """Returns all the objects."""
        return self._objects
